from fpylib.composer import compose, parallelize
from fpylib.lazyness import lazy_eval, lazy_class
from fpylib.irange import irange, char_encode, char_decode, num_irange, char_irange
from fpylib.functors import *
